import logging
import json

from collections import namedtuple
from typing import List, Dict, Union, Optional

import requests
from requests.compat import urljoin  # type: ignore
from requests.utils import prepend_scheme_if_needed

from ppa import appliance
from ppa_appliance import vaults

from . import exceptions

logger = logging.getLogger(__name__)


def _assert_engine_version(engine_version):
    if engine_version not in {1, 2}:
        raise exceptions.InvalidEngineVersion(f"The kv_version must be 1 or 2. Received {engine_version}.")


def proxies():
    return {"https": appliance.api_proxy(), "http": appliance.api_proxy()}


def verify():
    return appliance.cert()


def _get_vault(reason: str, vault_name: Optional[str] = None) -> vaults.Vault:
    return (
        vaults.hashicorp_vault_by_name(vault_name, reason=reason)
        if vault_name
        else vaults.select_vault(vaults.hashicorp_vaults(reason=reason))
    )


def _form_secret_url(vault_url: str, secret: str, engine: str, kv_version: int) -> str:
    return urljoin(
        prepend_scheme_if_needed(vault_url, "https"),
        f"/v1/{engine}/{secret}" if kv_version == 1 else f"/v1/{engine}/data/{secret}",
    )


def _parse_response(response: requests.Response) -> dict:
    try:
        return response.json()
    except json.decoder.JSONDecodeError:
        message = f"could not convert the following response body to JSON: {response.text}"
        logger.error(message)
        raise exceptions.VaultError(message)


def _unpack_and_raise(action: str, secret: str, engine: str, response: requests.Response, kv_version: int) -> None:
    status_code = response.status_code
    response_json = _parse_response(response)
    logger.debug(f"{response_json=}")
    combined_errors = ", ".join((response_json.get("warnings") or []) + (response_json.get("errors") or []))
    logger.error(f"status code: {status_code}, combined errors: {combined_errors}")

    if status_code == 404:
        if "no handler for route" in combined_errors:
            raise exceptions.InvalidEngine(
                f"Failed to {action} secret data. Supplied v{kv_version} engine '{engine}' does not exist."
            )
        elif "Invalid path for a versioned K/V secrets engine" in combined_errors:
            raise exceptions.InvalidSecret(
                f"Failed to {action} secret data. Supplied secret '{secret}' exists but is held in a KV "
                f"engine of a different version. The operation was attempted using the v{kv_version} API."
            )
        elif not combined_errors:
            raise exceptions.InvalidSecret(
                f"Failed to {action} secret data. Supplied secret '{secret}' does not "
                f"exist in v{kv_version} engine '{engine}'."
            )

    if status_code == 503 and "Vault is sealed" in combined_errors:
        raise exceptions.VaultSealed(f"Failed to {action} secret data. The Hashicorp Vault is sealed.")

    raise exceptions.VaultError(
        f"Failed to {action} keys against Hashicorp Vault secret '{secret}' using v{kv_version} engine "
        f"'{engine}'. Status code {status_code}. Please view logs for more information."
    )


def _filter_keys(data: Dict[str, str], keys: List[str], secret: str) -> Dict[str, str]:
    """
    - Takes a secret dictionary and a list of keys.
    - Returns a dictionary that only contains the requested keys.
    """
    return_data = {}
    for key in keys:
        try:
            return_data[key] = data[key]
        except KeyError:
            raise exceptions.InvalidKey(
                f"Failed to read secret data. Requested key '{key}' does not exist in secret '{secret}'."
            )
    return return_data


def _list_secret(url: str) -> requests.Response:
    """ List all keys and values from a Hashicorp Vault secret."""
    logger.info(f"listing secret using url {url}")
    return requests.get(url, proxies=proxies(), verify=verify())


def _update_secret(url: str, data: Dict[str, str]) -> requests.Response:
    """ Update a Hashicorp Vault secret with a new set of data.
    """
    logger.debug(f"sending post request to {url}")
    return requests.post(url, proxies=proxies(), verify=verify(), data=json.dumps(data))


def _load_secret_record(data: dict):
    """ We don't know which fields will exist in each secret until either we list the secret, or the user
    supplies the keys. For that reason we have to create a namedtuple based class on the fly.
    """
    fields = list(data.keys())
    logger.debug(f"creating VaultSecret record with fields {fields}")

    # Silence the following mypy error:
    # 'error: List or tuple literal expected as the second argument to namedtuple()'
    # There's no way I'm defining a list twice just to make mypy happy.
    VaultSecret = namedtuple("HashicorpVaultSecret", fields)  # type: ignore

    # Silence the following mypy error 'error: Too many arguments for "VaultSecret"'
    return VaultSecret(*[data.get(field) for field in fields])  # type: ignore


def _read(
    *,
    secret: str,
    engine: str,
    keys: Optional[List[str]] = None,
    reason: Optional[str] = "Task is accessing keys from Hashicorp Vault",
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
):
    """ List a Hashicorp Vault secret and return only the values for the supplied keys.
    """
    _assert_engine_version(kv_version)

    logger.info(f"attempting to read keys from secret {engine}/{secret}")
    hashicorp_vault = _get_vault(reason, vault_name=vault_name)
    response = _list_secret(_form_secret_url(hashicorp_vault.url, secret, engine, kv_version))

    if response.status_code != 200:
        _unpack_and_raise("read", secret, engine, response, kv_version)

    response_json = _parse_response(response)

    try:
        records = response_json["data"] if kv_version == 1 else response_json["data"]["data"]
        assert records
    except (KeyError, AssertionError):
        raise exceptions.VaultError(
            f"Hashicorp Vault secret '{secret}' in engine '{engine}' exists but contains no data."
        )
    return _load_secret_record(_filter_keys(records, keys, secret) if keys else records)


def _update(
    *,
    secret: str,
    engine: str,
    data: Dict[str, str],
    reason: Optional[str] = None,
    vault_name: Optional[str] = None,
    kv_version: int,
) -> None:
    """ Update one or more keys in a Hashicorp Vault secret.
    """
    _assert_engine_version(kv_version)

    def _get_existing(url) -> Union[Dict[str, str], None]:
        response_json = _parse_response(_list_secret(url))
        secret_record = (
            response_json.get("data", {}) if kv_version == 1 else response_json.get("data", {}).get("data", {})
        )
        logger.debug(f"secret has {len(secret_record)} existing keys that will be copied to the new secret")
        return secret_record

    def _build_update_data(master_record: dict, new_records: Dict[str, str]) -> dict:
        for key, value in new_records.items():
            logger.debug(f"adding supplied value for {key} into secret payload")
            master_record[key] = value
        return master_record if kv_version == 1 else {"data": master_record}

    update_count = len(data)
    hashicorp_vault = _get_vault(
        reason or f"Task is updating {update_count} keys in Hashicorp Vault", vault_name=vault_name
    )
    secret_url = _form_secret_url(hashicorp_vault.url, secret, engine, kv_version)
    logger.info(f"attempting to update {update_count} keys in v{kv_version} engine using path {secret_url}")

    response = _update_secret(secret_url, _build_update_data(_get_existing(secret_url), data))

    if any([all([kv_version == 1, response.status_code != 204]), all([kv_version == 2, response.status_code != 200])]):
        _unpack_and_raise("update", secret, engine, response, kv_version)
