"""Key-Value Engine

This module contains actions for reading & updating secrets in Key-Value engines.

!!! question "Supported Versions"

    This plugin supports Hashicorp Vault Key-Value versions 1 & 2.

!!! danger "Remember"

    You must have a Hashicorp Vault provisioned as a _Vault_ inside PPA to use this plugin.
"""

from typing import Dict, List, Optional, Any

from hashicorp_vault import kv_engine


def read_secret(
    secret: str,
    engine: Optional[str] = "kv",
    reason: Optional[str] = None,
    keys: Optional[List[str]] = None,
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
) -> Dict[str, Any]:
    """Read the values of a secret in a Key-Value engine.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __engine:__ the name of the secrets engine (defaults to `kv`)

    - __secret:__ the name of the secret in the secrets engine

    - __keys:__ a list of key names to get from the secret (defaults to all keys)

    - __reason:__ an audit message for the task interface (defaults to _Task is accessing keys from Hashicorp Vault_)

    - __vault_name:__ the Hashicorp Vault name in PPA (only necessary if more than one vault is configured)

    - __kv_version:__ the version of the key value engine (defaults to `2`)

??? success "Output"
    A dictionary containing the keys and values from the secret.

??? example

    - Getting all keys from the _active_directory_ secret in the _production_ engine

    - Saving them as a new variable called `active_directory_secrets`.

    ```YAML linenums="1"
    - hashicorp_vault.key_value.read_secret:
        secret: active_directory
        engine: production
      save: active_directory_secrets
    ```
"""
    return kv_engine.read_secret(
        engine=engine, secret=secret, keys=keys, reason=reason, vault_name=vault_name, kv_version=kv_version
    )._asdict()


def update_secret(
    secret: str,
    data: Dict[str, Any],
    engine: Optional[str] = "kv",
    reason: Optional[str] = None,
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
) -> None:
    """Update the values of a secret in a Key-Value engine.

This action can be used to:

- Replace values in a secret
- Append new values to a secret
- Create a new secret

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __engine:__ the name of the secrets engine (defaults to `kv`)

    - __secret:__ the name of the secret in the secrets engine

    - __data:__ a dictionary containing keys & values to update in the secret

    - __reason:__ an audit message for the task interface (defaults to _Task is accessing keys from Hashicorp Vault_)

    - __vault_name:__ the Hashicorp Vault name in PPA (only necessary if more than one vault is configured)

    - __kv_version:__ the version of the key value engine (defaults to `2`)

??? success "Output"
    Nothing is outputted by this action.

??? example

    Updating the _service_now_ secret's _access_token_ value in the _production_ engine.

    ```YAML linenums="1"
    - hashicorp_vault.key_value.update_secret:
        secret: service_now
        engine: production
      load:
        data:
          access_token: new_access_token
    ```
    """
    return kv_engine.read_secret(
        engine=engine, secret=secret, data=data, reason=reason, vault_name=vault_name, kv_version=kv_version
    )
